#ifndef _DOUBLY_LINKED_LIST_ITERATOR_H_
#define _DOUBLY_LINKED_LIST_ITERATOR_H_


namespace SCE {


/*
 * Structure used to staticly (at compile time) choose the insertion method 
 * used in the linked list. This is the structure used when ordered == false.
 */
template<class Container, class T, bool ordered>
struct DL_SelectInsertionType
{
  static T const & add(Container * & current_node, T const & element)
        {
		// create a new node
		Container * new_node = new Container;
		
		// save a copy of element in the new node
		new_node->element = element;

		// insert the new node in the list
		new_node->previous = current_node;
		new_node->next = current_node->next;
		current_node->next = new_node;

		// make current_node point to the new node
		current_node = new_node;
	
		// return the reference to the element
                return element;
        }
};

/*
 * Structure used to staticly (at compile time) choose the insertion method 
 * used in the linked list. This is the structure used when ordered == true.
 */
template<class Container, class T>
struct DL_SelectInsertionType <Container, T, true>
{
        static T const &
	add(Container * & current_node, T const & element)
        {
		// create a new node
		Container * new_node = new Container;
		
		// save a copy of element in the new node
		new_node->element = element;

		// if the list isn't empty
		if (current_node->next != NULL && current_node->previous != NULL)
		{
			// search for the right position in the linked list for
			// the new node
			//
			// new elements will be inserted in order of size
			// (operator >) and in order of arrival if there is
			// a draw
			if (current_node->element <= element)
			{ 
                		while (current_node->next != NULL && 
					current_node->next->element <= element)
                		{
                        		current_node = current_node->next;
                		}
			}
			else
			{
                		while (current_node->previous != NULL && 
					current_node->element > element)
                		{
                        		current_node = current_node->previous;
                		}
			}
		}

		// insert the new node in the list
		new_node->previous = current_node;
		new_node->next = current_node->next;
		current_node->next = new_node;

		// make current_node point to the new node
		current_node = new_node;
			
		// return the reference to the element	
                return element;
        }
};

/*
 * DoublyLinkedListIterator class.
 */
template <class Container, class T, bool ordered>
class DoublyLinkedListIterator
{
private:
	// a pointer to the node this iterator currentlu points to
	// current_node also give the current address of the iterator
	Container * current_node;

public:
	// constructor
	DoublyLinkedListIterator()
	{
		current_node = NULL;
	};

	// constructor
	DoublyLinkedListIterator(Container * node)
	{
		current_node = node;
	}

	// destructor
	~DoublyLinkedListIterator()
	{
		// empty
	};


	// returns false if there's at least one element in the list 
        bool isEmpty() const
	{
		return current_node->next == NULL &&
			current_node->previous == NULL;
	}

	// returns true if current_node isn't the last node in the list
        bool hasNext() const
	{
		return current_node->next != NULL;
	}

	// returns true if current_node isn't the first (or the head) node of
	// the list
        bool hasPrevious() const
	{
		// if current_node isn't the head of the list
		if (current_node->previous != NULL)
		{
			return current_node->previous->previous != NULL;
		}
		return false;
	}

	// remove the current_node form the linked list
        void remove()
	{
		// if the list isn't empty
		if (!isEmpty())
		{
			// if current_node isn't the head of the list
			if (current_node->previous != NULL)
			{
				// saving a reference to current_node
				Container * tmp = current_node;

				// removing current node from the linke list
				current_node->previous->next = current_node->next;

				// if current_node isn't the tail of the list
				if (hasNext())
				{
					// finishing the removal of current_node
					current_node->next->previous = 
						current_node->previous;
					// make current_node point to the next
					// node in the list
					current_node = current_node->next;
				}
				else
				{
					// make current_node point to the
					// previous node in the list
					current_node = current_node->previous;
				}

				// deleting the node
				delete tmp;
			}
			else  // current_node is the head_node
			{
				// make current_node point to the first node
				current_node = current_node->next;

				// and remove it (current_node will point to the				// node right after the one removed if it exists
				// or to the head_node otherwise)
				remove();
			}
		}
	}


	// equals to operator
        bool operator==(DoublyLinkedListIterator<Container, T, ordered> & iterator) const
        {
		// compares the addresses of two iterators
		return current_node == &iterator;
	}

	// differs from operator
        bool operator!=(DoublyLinkedListIterator<Container, T, ordered> & iterator) const
        {
		// compares the addresses of two iterators
		return current_node != &iterator;
	}

	// address-of operator
	Container * const operator&() const
	{
		// return the address in current_node
		return current_node;
	}

	// assignement operator
        T const & operator=(T const & value)
        {
		// call the proper function (depending on the ordered parameter)
		return DL_SelectInsertionType<Container, T, ordered>::
			add(current_node, value);
        }

	// dereference operator
	T & operator*() const
	{
		// returns the element in current_node
		return current_node->element;
        }

	// prefix ++
	Container * const operator++()
	{
		// make current_node point to the next node in the list (if any)
		if (current_node->next != NULL)
		{
			current_node = current_node->next;
		}

		// return the address in current_node
		return current_node; 
	}

	// postfix ++
	Container * const operator++(int)
	{
		// make current_node point to the next node in the list (if any)
		// and return the address of the previous node
		if (current_node->next != NULL)
		{
			current_node = current_node->next;
			return current_node->previous;
		}
		// return the address in current_node
		return current_node;
	}

	// prefix --
	Container * const & operator--()
	{
		// make current_node point to the previous node (if any)
		if (current_node->previous != NULL)
		{
			current_node = current_node->previous;
		}
		// return the address in current_node
		return current_node; 
	}

	// postfix --
	Container * const operator--(int)
	{
		// make current_node point to the previous node in the list
		// (if any) and return the address of the node
		if (current_node->previous != NULL)
		{
			current_node = current_node->previous;
			return current_node->next;
		}
		// return the address in current_node
		return current_node;
	}
};

}  // namespace SCE

#endif  /* _DOUBLY_LINKED_LIST_ITERATOR_H_ */
