#ifndef _SINGLE_LINKED_LIST_ITERATOR_H_
#define _SINGLE_LINKED_LIST_ITERATOR_H_


namespace SCE {

	
/*
 * Structure used to staticly (at compile time) choose the insertion method 
 * used in the linked list. This is the structure used when ordered == false.
 */
template<class Container, class T, bool ordered>
struct SL_SelectInsertionType
{
	static T const &
	add(Container * const & head_node, Container * & current_node, T const & element)
	{
		// create a new node
		Container * new_node = new Container;
		
		// save a copy of the node next to current_node
		Container * tmp = current_node->next;
	
		// save a copy of element in the new node
		new_node->element = element;

		// insert the new node in the list
		current_node->next = new_node;
		new_node->next = tmp;

		// make current_node point to the new node
		current_node = new_node;

		// return the reference to the element
		return element;
	}
};

/*
 * Structure used to staticly (at compile time) choose the insertion method 
 * used in the linked list. This is the structure used when ordered == true.
 */
template<class Container, class T>
struct SL_SelectInsertionType <Container, T, true>
{
	static T const &
	add(Container * const & head_node, Container * & current_node, T const & element)
	{
		// create a new node
		Container * new_node = new Container;

		// save a copy of element in the new node
		new_node->element = element;

		// run thru the list until you find the proper spot for the new
		// node
		//
		// new elements will be inserted in order of size (operator >) 
		// and in order of arrival if there's a draw
		Container * tmp = head_node;

		while(tmp->next != NULL && tmp->next->element <= element)
		{
			tmp = tmp->next;
		}
	
		current_node = tmp;

		tmp = current_node->next;

		// insert the new node in the list
		current_node->next = new_node;
		new_node->next = tmp;

		// make current_node point to the new node
		current_node = new_node;

		// return the reference to the element
		return element;
	}
};


/*
 * SingleLinkedListIterator class.
 */
template <class Container, class T, bool ordered>
class SingleLinkedListIterator
{
private:
	// a pointer to the head node in the list
	Container * head_node;
	
	// a pointer to the node this iterator currently points to
        // current_node also give the current address of the iterator
	Container * current_node; 

public:
	// constructor
	SingleLinkedListIterator()
	{
		head_node = NULL;
		current_node = NULL;
	};
	
	// constructor
	SingleLinkedListIterator(Container * node)
	{
		head_node = node;
		current_node = head_node;
	}
	
	// destructor
	~SingleLinkedListIterator()
	{
		// empty
	};


	// returns false if there's at least one element in the list 
        bool isEmpty() const
	{
		return head_node->next == NULL;
	}

	// returns true if current_node isn't the last node in the list	
        bool hasNext() const
	{
		return current_node->next != NULL;
	}
	
	// remove current_node from the linked list
        void remove()
	{
		// if the list isn't empty
		if (!isEmpty())
		{
			// if current_node isn't the head of the list
			if (current_node != head_node)
			{
				Container * tmp = head_node;

				// run thru the linked list until you find
				// current_node
				while(tmp->next != current_node)
				{
					tmp = tmp->next;
				}

				// remove current node from the linke list
				tmp->next = current_node->next;
				delete current_node;

				// make current_node point to the next node 
				// in the list if it exists
				if (tmp->next != NULL)
				{
					current_node = tmp->next;
	
				}
				else  // or the the previous one otherwise
				{
					current_node = tmp;
				}
			}
			else  // current_node is the head_node
			{
				// make current_node point to the first node
				current_node = head_node->next;

				// and remove it (current_node will point to the
				// node right after the one removed if it exists
				// or to the head_node otherwise)
				remove();
			}

		}
	}	


	// equals to operator
        bool operator==(SingleLinkedListIterator<Container, T, ordered> & iterator) const
        {
		// compares the addresses of two iterators
		return current_node == &iterator;
	}

	// differs from operator
        bool operator!=(SingleLinkedListIterator<Container, T, ordered> & iterator) const
        {
		// compares the addresses of two iterators
		return current_node != &iterator;
	}

	// address-of operator
	Container * const operator&() const
	{
		// return the address in current_node
		return current_node;
	}

	// assignment operator
        T const & operator=(T const & element)
        {
		// call the proper function (depending on the ordered parameter)
		return SL_SelectInsertionType<Container, T, ordered>::
			add(head_node, current_node, element); 
        }

	// dereference operator
	T & operator*() const
	{
		// returns the element in current_node
		return current_node->element;
        }

	// prefix ++
	Container * const operator++()
	{
		// make current_node point to the next node in the list (if any)
		if (current_node->next != NULL)
		{
			current_node = current_node->next;
		}
		// return the address in current_node
		return current_node; 
	}
	
	// postfix ++
	Container * const operator++(int)
	{
		// make current_node point to the next node in the list (if any)
		// and return the address of the previous node
		if (current_node->next != NULL)
		{
			Container * tmp = current_node;
			current_node = current_node->next;
			return tmp;
		}
		// return the address in current_node
		return current_node;
	}
};

}  // namespace SCE

#endif  /* _SINGLE_LINKED_LIST_ITERATOR_H_ */
